(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Parser_Monad_Unused
imports
  "$AFP/Show/Show"
  Error_Monad
begin

datatype_new ('t, 'a) gen_parser =
  Parser "'t list \<Rightarrow> ((string + 'a) \<times> 't list) option"

type_synonym
  'a parser = "(char, 'a) gen_parser"

text {*Allow extensional view on parsers by extracting the underlying function.*}
fun
  run :: "('t, 'a) gen_parser \<Rightarrow> 't list \<Rightarrow> ((string + 'a) \<times> 't list) option"
where
  "run (Parser p) = p"

definition
  parse :: "('t, 'a) gen_parser \<Rightarrow> 't list \<Rightarrow> string + 'a"
where
  "parse p ts =
    (case run p ts of
      Some (r, _) \<Rightarrow> r
    | None \<Rightarrow> error ''parse: unkown error'')"

lemma parser_cases [case_names succeed fail]:
  fixes p and ts
  assumes succeed: "\<And>x ts'. run p ts = Some (x, ts') \<Longrightarrow> P"
    and fail: "run p ts = None \<Longrightarrow> P"
  shows P
  using assms by (cases "run p ts") auto

lemma Parser_run [simp]:
  "Parser (run p) = p"
  by (cases p) simp

lemma parser_eqI [intro]:
  "(\<And>ts. run p ts = run q ts) \<Longrightarrow> p = q"
  by (cases p, cases q) (auto simp: fun_eq_iff)

ML {*
structure Run_Simps = Named_Thms (
  val name = @{binding "run_simps"}
  val description = "simplification rules for run"
)
*}

setup Run_Simps.setup

lemma run_Let [run_simps]:
  "run (let x = t in f x) = (let x = t in run (f x))"
  by (simp add: Let_def)

definition
  parser :: "('t list \<Rightarrow> (string + 'a) \<times> 't list) \<Rightarrow> ('t, 'a) gen_parser"
where
  "parser f = Parser (Some o f)"

lemma run_parser [run_simps]:
  "run (parser f) = Some o f"
  by (simp add: parser_def)

definition
  return :: "'a \<Rightarrow> ('t, 'a) gen_parser"
where
  "return x = parser (Pair (Error_Monad.return x))"

lemma run_return [run_simps]:
  "run (return x) = Some o Pair (Inr x)"
  by (simp add: return_def run_simps)

definition
  error :: "string \<Rightarrow> ('t, 'a) gen_parser"
where
  "error e = parser (Pair (Error_Monad.error e))"

text {*Unrecoverable errors (needed for @{text partial_function}).*}
definition
  fail :: "('t, 'a) gen_parser"
where
  [code del]: "fail = Parser (\<lambda>_. None)"

lemma run_error [run_simps]:
  "run (error s) = Some o Pair (Inl s)"
  by (simp add: error_def run_simps)

lemma run_fail [run_simps]:
  "run fail = (\<lambda>_. None)"
  by (simp add: fail_def)
  
definition
  bind :: "('t, 'a) gen_parser \<Rightarrow> ('a \<Rightarrow> ('t, 'b) gen_parser) \<Rightarrow> ('t, 'b) gen_parser"
where
  "bind p f = Parser (\<lambda>ts.
    (case run p ts of
      Some (Inr x, ts') \<Rightarrow> run (f x) ts'
    | Some (Inl e, ts') \<Rightarrow> Some (Inl e, ts')
    | None \<Rightarrow> None))"

adhoc_overloading
  Monad_Syntax.bind bind

lemma bind_cong [fundef_cong]:
  fixes m1 :: "('t, 'a) gen_parser"
  assumes "run m1 ts2 = run m2 ts2"
    and "\<And>y ts. run m2 ts2 = Some (Inr y, ts) \<Longrightarrow> run (f1 y) ts = run (f2 y) ts"
    and "ts1 = ts2"
  shows "run (m1 \<guillemotright>= f1) ts1 = run (m2 \<guillemotright>= f2) ts2"
  using assms unfolding bind_def
  by (cases "run m1 ts1") (auto split: sum.splits)

definition
  update_tokens :: "('t list \<Rightarrow> 't list) \<Rightarrow> ('t, 't list) gen_parser"
where
  "update_tokens f = parser (\<lambda>ts. (Error_Monad.return ts, f ts))"

definition
  get_tokens :: "('t, 't list) gen_parser"
where
  "get_tokens = update_tokens (\<lambda>x. x)"

definition
  set_tokens :: "'t list \<Rightarrow> ('t, unit) gen_parser"
where
  [code_unfold]: "set_tokens ts = update_tokens (\<lambda>_. ts) \<guillemotright> return ()"

definition
  err_expecting :: "string \<Rightarrow> ('t::show, 'a) gen_parser"
where
  "err_expecting msg = parser (\<lambda>ts.
    (Error_Monad.error (''expecting '' @ msg @ '', but found: '' @
      shows_quote (shows (take 30 ts)) []), ts))"

definition eoi :: "('t::show, unit) gen_parser" where
  "eoi = do {
    ts \<leftarrow> get_tokens;
    if ts = [] then return ()
    else err_expecting ''end of input''
  }"

lemma run_bind [run_simps]:
  "run p ts = Some (Inr x, ts') \<Longrightarrow> run (p \<guillemotright>= f) ts = run (f x) ts'"
  "run p ts = Some (Inl e, ts') \<Longrightarrow> run (p \<guillemotright>= f) ts = Some (Inl e, ts')"
  "run p ts = None \<Longrightarrow> run (p \<guillemotright>= f) ts = None"
  by (simp_all add: bind_def)

lemma run_bind_case:
  "run (p \<guillemotright>= f) ts =
    (case run p ts of
      Some (Inr x, ts') \<Rightarrow> run (f x) ts'
    | Some (Inl e, ts') \<Rightarrow> Some (Inl e, ts')
    | None \<Rightarrow> None)"
  by (simp add: bind_def)


subsection {* Monad Laws *}

lemma return_bind [simp]:
  "(return x \<guillemotright>= f) = f x"
  by (rule parser_eqI) (simp add: run_simps)

lemma bind_return [simp]:
  "(p \<guillemotright>= return) = p"
  by (rule parser_eqI) (simp add: bind_def run_simps split: option.splits sum.splits)

lemma bind_assoc [simp]:
  "(p \<guillemotright>= f) \<guillemotright>= g = (p :: ('t, 'a) gen_parser) \<guillemotright>= (\<lambda>x. f x \<guillemotright>= g)"
  by (rule parser_eqI) (simp add: bind_def run_simps split: option.splits sum.splits)

lemma error_bind [simp]:
  "(error e \<guillemotright>= f) = error e" 
  by (rule parser_eqI) (simp add: run_simps)

lemma fail_bind [simp]:
  "(fail \<guillemotright>= f) = fail"
  by (rule parser_eqI) (simp add: run_simps)


subsection {* Basic Parsers *}

definition
  sat :: "('t \<Rightarrow> bool) \<Rightarrow> ('t, 't) gen_parser"
where
  "sat p = Parser (\<lambda>ts.
    (case ts of
      [] \<Rightarrow> Some (Inl ''sat: empty input'', [])
    | t#ts \<Rightarrow> Some
      (if p t then (Inr t, ts)
      else (Inl ''sat: wrong token'', t#ts))))"


subsection {* Basic Combinators *}

definition
  orelse :: "('t, 'a) gen_parser \<Rightarrow> ('t, 'a) gen_parser \<Rightarrow> ('t, 'a) gen_parser" (infixr "<|>" 60)
where
  "p <|> q = Parser (\<lambda>ts.
    (case run p ts of
      Some (Inr x, ts') \<Rightarrow> Some (Inr x, ts')
    | Some (Inl e, ts') \<Rightarrow> run q ts
    | None \<Rightarrow> None))"

lemma run_orelse [run_simps]:
  "run p ts = Some (Inr x, ts') \<Longrightarrow> run (p <|> q) ts = Some (Inr x, ts')"
  "run p ts = Some (Inl e, ts') \<Longrightarrow> run (p <|> q) ts = run q ts"
  "run p ts = None \<Longrightarrow> run (p <|> q) ts = None"
  by (simp_all add: orelse_def)

lemma run_orelse_case:
  "run (p <|> q) ts =
    (case run p ts of
      Some (Inr x, ts') \<Rightarrow> Some (Inr x, ts')
    | Some (Inl e, ts') \<Rightarrow> run q ts
    | None \<Rightarrow> None)"
  by (simp add: orelse_def)


subsection {* Partial function definition setup *}

definition
  parser_ord :: "('t, 'a) gen_parser \<Rightarrow> ('t, 'a) gen_parser \<Rightarrow> bool"
where
  "parser_ord = img_ord run (fun_ord option_ord)"

definition
  parser_lub :: "('t, 'a) gen_parser set \<Rightarrow> ('t, 'a) gen_parser"
where
  "parser_lub = img_lub run Parser (fun_lub (flat_lub None))"

interpretation parser!: partial_function_definitions parser_ord parser_lub
proof -
  have "partial_function_definitions (fun_ord option_ord) (fun_lub (flat_lub None))"
    by (rule partial_function_lift) (rule flat_interpretation)
  hence "partial_function_definitions (img_ord run (fun_ord option_ord))
    (img_lub run Parser (fun_lub (flat_lub None)))"
    by (rule partial_function_image) (auto)
  thus "partial_function_definitions parser_ord parser_lub"
    by (simp only: parser_ord_def parser_lub_def)
qed

declaration {*
  Partial_Function.init
    "parser"
    @{term parser.fixp_fun}
    @{term parser.mono_body}
    @{thm parser.fixp_rule_uc}
    @{thm parser.fixp_induct_uc}
    NONE
*}

abbreviation "mono_parser \<equiv> monotone (fun_ord parser_ord) parser_ord"

lemma parser_ordI:
  assumes "\<And>ts. run p ts = None \<or> run p ts = run q ts"
  shows "parser_ord p q"
  using assms unfolding parser_ord_def img_ord_def fun_ord_def flat_ord_def
  by blast

lemma parser_ordE:
  assumes "parser_ord p q"
  obtains "run p ts = None" | "run p ts = run q ts"
  using assms unfolding parser_ord_def img_ord_def fun_ord_def flat_ord_def
  by atomize_elim blast

lemma bind_mono [partial_function_mono]:
  assumes mf: "mono_parser (B :: ('a \<Rightarrow> ('t, 'b) gen_parser) \<Rightarrow> ('t, 'c) gen_parser)"
    and mg: "\<And>y. mono_parser (\<lambda>f. C y f)"
  shows "mono_parser (\<lambda>f. B f \<guillemotright>= (\<lambda>y. C y f))"
proof (rule monotoneI)
  fix f g :: "'a \<Rightarrow> ('t, 'b) gen_parser"
  assume fg: "fun_ord parser_ord f g"
  from mf have 1: "parser_ord (B f) (B g)" by (rule monotoneD) (rule fg)
  from mg have 2: "\<And>y'. parser_ord (C y' f) (C y' g)" by (rule monotoneD) (rule fg)
  have "parser_ord (B f \<guillemotright>= (\<lambda>y. C y f)) (B g \<guillemotright>= (\<lambda>y. C y f))" (is "parser_ord ?L ?R")
  proof (rule parser_ordI)
    fix ts
    from 1 show "run ?L ts = None \<or> run ?L ts = run ?R ts"
      by (rule parser_ordE [where ts = ts]) (auto simp: run_bind_case)
  qed
  also have "parser_ord (B g \<guillemotright>= (\<lambda>y'. C y' f)) (B g \<guillemotright>= (\<lambda>y'. C y' g))" (is "parser_ord ?L ?R")
  proof (rule parser_ordI)
    fix ts
    show "run ?L ts = None \<or> run ?L ts = run ?R ts"
    proof (cases "run (B g) ts")
      case None
      hence "run ?L ts = None" by (auto simp: run_bind_case)
      thus ?thesis ..
    next
      case Some
      then obtain r ts' where 3: "run (B g) ts = Some (r, ts')"
        by (metis surjective_pairing)
      show ?thesis
      proof (cases r)
        case (Inr x)
        from 3 have "run ?L ts = run (C x f) ts'" and "run ?R ts = run (C x g) ts'"
          unfolding Inr by (auto simp: run_bind_case)
        with 2[of x] show ?thesis by (auto elim: parser_ordE)
      next
        case (Inl e)
        from 3 have "run ?L ts = Some (Inl e, ts')" and "run ?R ts = Some (Inl e, ts')"
          unfolding Inl by (auto simp: run_bind_case)
        thus ?thesis by simp
      qed
    qed
  qed
  finally (parser.leq_trans) show "parser_ord (B f \<guillemotright>= (\<lambda>y. C y f)) (B g \<guillemotright>= (\<lambda>y'. C y' g))" .
qed

lemma bind_orelse [partial_function_mono]:
  assumes mf: "mono_parser (A :: ('a \<Rightarrow> ('t, 'b) gen_parser) \<Rightarrow> ('t, 'c) gen_parser)"
    and mg: "mono_parser B"
  shows "mono_parser (\<lambda>f. A f <|> B f)"
proof (rule monotoneI)
  fix f g :: "'a \<Rightarrow> ('t, 'b) gen_parser"
  assume fg: "fun_ord parser_ord f g"
  from mf have 1: "parser_ord (A f) (A g)" by (rule monotoneD) (rule fg)
  from mg have 2: "parser_ord (B f) (B g)" by (rule monotoneD) (rule fg)
  show "parser_ord (A f <|> B f) (A g <|> B g)" (is "parser_ord ?L ?R")
  proof (rule parser_ordI)
    fix ts
    from 1 show "run ?L ts = None \<or> run ?L ts = run ?R ts"
    proof (rule parser_ordE)
      assume "run (A f) ts = None" thus ?thesis by (auto simp: run_orelse_case)
    next
      assume eq: "run (A f) ts = run (A g) ts"
      show ?thesis
      proof (cases "run (A f) ts")
        case None thus ?thesis by (auto simp: run_orelse_case)
      next
        case (Some result)
        show ?thesis
        proof (cases result)
          case (Pair r ts')
          show ?thesis
          proof (cases r)
            case (Inr x)
            have 3: "run (A f <|> B f) ts = run (A f) ts"
             by (auto simp: Some Pair Inr run_orelse_case)
            have 4: "run (A g <|> B g) ts = run (A f) ts"
              using run_orelse(1)[of "A g" ts x ts'] and eq
              unfolding Some Pair Inr by simp
            show ?thesis by (simp add: 3 4)
          next
            case (Inl e)
            have 3: "run (A f <|> B f) ts = run (B f) ts"
              by (auto simp: Some Pair Inl run_orelse_case)
            have 4: "run (A g <|> B g) ts = run (B g) ts"
              using run_orelse(2)[of "A g" ts e ts'] and eq
              unfolding Some Pair Inl by simp
            from 2
            show ?thesis
            proof (rule parser_ordE)
              assume "run (B f) ts = None" thus ?thesis by (simp add: 3 4)
            next
              assume "run (B f) ts = run (B g) ts" thus ?thesis by (simp add: 3 4)
            qed
          qed
        qed
      qed
    qed
  qed
qed


subsection {* Derived Parsers *}

definition any :: "('t, 't) gen_parser" where
  "any = sat (\<lambda>_. True)"

definition oneof :: "'t list \<Rightarrow> ('t, 't) gen_parser" where
  "oneof ts = sat (\<lambda>t. t \<in> set ts)"

definition noneof :: "'t list \<Rightarrow> ('t, 't) gen_parser" where
  "noneof ts = sat (\<lambda>t. t \<notin> set ts)"


subsection {* Derived Combinators *}

partial_function (parser)
  many :: "('t, 'a) gen_parser \<Rightarrow> ('t, 'a list) gen_parser"
where
  "many p = ((do { x \<leftarrow> p; xs \<leftarrow> many p; return (x#xs) }) <|> return [])"
declare many.simps [code]

definition
  many1 :: "('t, 'a) gen_parser \<Rightarrow> ('t, 'a list) gen_parser"
where
  "many1 p = do { x \<leftarrow> p; xs \<leftarrow> many p; return (x#xs) }"

export_code
  (* parsers *)
  any error noneof oneof return sat
  (* combinators *)
  bind many many1 orelse
  (* auxiliary *)
  parse
  in Haskell module_name ParserMonad

end

